@@ -11,6 +11,7 @@ import time
 
 from .event import event
 from .messages import ExpanderMessage
+from .panels import ADEMCO, DSC
 
 
 class Zone(object):
@@ -37,8 +38,10 @@ class Zone(object):
     """Zone status"""
     timestamp = None
     """Timestamp of last update"""
+    expander = False
+    """Does this zone exist on an expander?"""
 
-    def __init__(self, zone=0, name='', status=CLEAR):
+    def __init__(self, zone=0, name='', status=CLEAR, expander=False):
         """
         Constructor
 
@@ -53,6 +56,7 @@ class Zone(object):
         self.name = name
         self.status = status
         self.timestamp = time.time()
+        self.expander = expander
 
     def __str__(self):
         """
@@ -116,7 +120,7 @@ class Zonetracker(object):
         """
         self._zones_faulted = value
 
-    def __init__(self):
+    def __init__(self, alarmdecoder_object):
         """
         Constructor
         """
@@ -124,6 +128,8 @@ class Zonetracker(object):
         self._zones_faulted = []
         self._last_zone_fault = 0
 
+        self.alarmdecoder_object = alarmdecoder_object
+
     def update(self, message):
         """
         Update zone statuses based on the current message.
@@ -132,9 +138,12 @@ class Zonetracker(object):
         :type message: :py:class:`~alarmdecoder.messages.Message` or :py:class:`~alarmdecoder.messages.ExpanderMessage`
         """
         if isinstance(message, ExpanderMessage):
-            if message.type == ExpanderMessage.ZONE:
-                zone = self.expander_to_zone(message.address, message.channel)
+            zone = -1
 
+            if message.type == ExpanderMessage.ZONE:
+                zone = self.expander_to_zone(message.address, message.channel, self.alarmdecoder_object.mode)
+
+            if zone != -1:
                 status = Zone.CLEAR
                 if message.value == 1:
                     status = Zone.FAULT
@@ -142,14 +151,12 @@ class Zonetracker(object):
                     status = Zone.CHECK
 
                 # NOTE: Expander zone faults are handled differently than
-                #       regular messages.  We don't include them in
-                #       self._zones_faulted because they are not reported
-                #       by the panel in it's rolling list of faults.
+                #       regular messages.
                 try:
                     self._update_zone(zone, status=status)
 
                 except IndexError:
-                    self._add_zone(zone, status=status)
+                    self._add_zone(zone, status=status, expander=True)
 
         else:
             # Panel is ready, restore all zones.
@@ -168,14 +175,8 @@ class Zonetracker(object):
                 self._last_zone_fault = 0
 
             # Process fault
-            elif message.check_zone or message.text.startswith("FAULT"):
-                # Apparently this representation can be both base 10
-                # or base 16, depending on where the message came
-                # from.
-                try:
-                    zone = int(message.numeric_code)
-                except ValueError:
-                    zone = int(message.numeric_code, 16)
+            elif self.alarmdecoder_object.mode != DSC and (message.check_zone or message.text.startswith("FAULT") or message.text.startswith("ALARM")):
+                zone = message.parse_numeric_code()
 
                 # NOTE: Odd case for ECP failures.  Apparently they report as
                 #       zone 191 (0xBF) regardless of whether or not the
@@ -195,6 +196,9 @@ class Zonetracker(object):
                     self._update_zone(zone)
                     self._clear_zones(zone)
 
+                    # Save our spot for the next message.
+                    self._last_zone_fault = zone
+
                 else:
                     status = Zone.FAULT
                     if message.check_zone:
@@ -204,12 +208,12 @@ class Zonetracker(object):
                     self._zones_faulted.append(zone)
                     self._zones_faulted.sort()
 
-                # Save our spot for the next message.
-                self._last_zone_fault = zone
+                    # A new zone fault, so it is out of sequence.
+                    self._last_zone_fault = 0
 
             self._clear_expired_zones()
 
-    def expander_to_zone(self, address, channel):
+    def expander_to_zone(self, address, channel, panel_type=ADEMCO):
         """
         Convert an address and channel into a zone number.
 
@@ -221,12 +225,19 @@ class Zonetracker(object):
         :returns: zone number associated with an address and channel
         """
 
-        # TODO: This is going to need to be reworked to support the larger
-        #       panels without fixed addressing on the expanders.
+        zone = -1
 
-        idx = address - 7   # Expanders start at address 7.
+        if panel_type == ADEMCO:
+            # TODO: This is going to need to be reworked to support the larger
+            #       panels without fixed addressing on the expanders.
 
-        return address + channel + (idx * 7) + 1
+            idx = address - 7   # Expanders start at address 7.
+            zone = address + channel + (idx * 7) + 1
+
+        elif panel_type == DSC:
+            zone = (address * 8) + channel
+
+        return zone
 
     def _clear_zones(self, zone):
         """
@@ -235,6 +246,11 @@ class Zonetracker(object):
         :param zone: current zone being processed
         :type zone: int
         """
+
+        if self._last_zone_fault == 0:
+            # We don't know what the last faulted zone was, nothing to do
+            return
+
         cleared_zones = []
         found_last_faulted = found_current = at_end = False
 
@@ -242,7 +258,7 @@ class Zonetracker(object):
         it = iter(self._zones_faulted)
         try:
             while not found_last_faulted:
-                z = it.next()
+                z = next(it)
 
                 if z == self._last_zone_fault:
                     found_last_faulted = True
@@ -255,7 +271,7 @@ class Zonetracker(object):
         # between to our clear list.
         try:
             while not at_end and not found_current:
-                z = it.next()
+                z = next(it)
 
                 if z == zone:
                     found_current = True
@@ -273,7 +289,7 @@ class Zonetracker(object):
 
             try:
                 while not found_current:
-                    z = it.next()
+                    z = next(it)
 
                     if z == zone:
                         found_current = True
@@ -286,7 +302,9 @@ class Zonetracker(object):
 
         # Actually remove the zones and trigger the restores.
         for z in cleared_zones:
-            self._update_zone(z, Zone.CLEAR)
+            # Don't clear expander zones, expander messages will fix this
+            if self._zones[z].expander is False:
+                self._update_zone(z, Zone.CLEAR)
 
     def _clear_expired_zones(self):
         """
@@ -294,14 +312,14 @@ class Zonetracker(object):
         """
         zones = []
 
-        for z in self._zones.keys():
+        for z in list(self._zones.keys()):
             zones += [z]
 
         for z in zones:
             if self._zones[z].status != Zone.CLEAR and self._zone_expired(z):
                 self._update_zone(z, Zone.CLEAR)
 
-    def _add_zone(self, zone, name='', status=Zone.CLEAR):
+    def _add_zone(self, zone, name='', status=Zone.CLEAR, expander=False):
         """
         Adds a zone to the internal zone list.
 
@@ -313,10 +331,9 @@ class Zonetracker(object):
         :type status: int
         """
         if not zone in self._zones:
-            self._zones[zone] = Zone(zone=zone, name=name, status=status)
+            self._zones[zone] = Zone(zone=zone, name=name, status=None, expander=expander)
 
-        if status != Zone.CLEAR:
-            self.on_fault(zone=zone)
+        self._update_zone(zone, status=status)
 
     def _update_zone(self, zone, status=None):
         """
@@ -332,9 +349,11 @@ class Zonetracker(object):
         if not zone in self._zones:
             raise IndexError('Zone does not exist and cannot be updated: %d', zone)
 
-        if status is not None:
-            self._zones[zone].status = status
+        old_status = self._zones[zone].status
+        if status is None:
+            status = old_status
 
+        self._zones[zone].status = status
         self._zones[zone].timestamp = time.time()
 
         if status == Zone.CLEAR:
@@ -342,6 +361,9 @@ class Zonetracker(object):
                 self._zones_faulted.remove(zone)
 
             self.on_restore(zone=zone)
+        else:
+            if old_status != status and status is not None:
+                self.on_fault(zone=zone)
 
     def _zone_expired(self, zone):
         """
@@ -352,4 +374,4 @@ class Zonetracker(object):
 
         :returns: whether or not the zone is expired
         """
-        return time.time() > self._zones[zone].timestamp + Zonetracker.EXPIRE
+        return (time.time() > self._zones[zone].timestamp + Zonetracker.EXPIRE) and self._zones[zone].expander is False
